function [samples, tail] = inference(documents, ...
                                      numSents, ...
                                      model, ...
                                      K, ...
                                      lmprior, ...
                                      tauprior, ...
                                      sigmaprior, ...
                                      rhoprior, ...
                                      nuprior, ...
                                      keep, ...
                                      iterations, ...
                                      savetime, ...
                                      filename)

%% ---------------------------------------------------------
% documents - vector(numDocs) of sparse vectors(vocab x sentence->count)
% numSents - number of sentences in each document
% K - topic count
% keep - number of last iterations to keep
% iterations - number of iterations
% model - 0 : constrained order model
%         1 : uniform order model
%         2 : Mallows model


rand('state', sum(100 * clock));
numDocs = size(numSents,1);

% Vocabulary size
vocabSize = size(documents(1).count, 1);
maxNumSents = max(numSents);

% Dirichlet prior values
theta_n_0 = lmprior;
theta_t_0 = lmprior;
theta_d_0 = lmprior;
tau_0 = tauprior;
sigma_0 = sigmaprior * ones(3, 1);
v_0 = vprior(rhoprior, K);
nu_0 = nuprior;

% Samples array
samples = struct('t',       cell(keep, 1), ...
                 's',       cell(keep, 1), ...
                 'v',       cell(keep, 1), ...
                 'rho',     cell(keep, 1));

% Tail of circular queue
tail = 1;

%% ---------------------------------------------------------
% These matrices keep track of counts for: word in topic, word in document,
% word in noise, topic in document, and word source in document
ntheta_t = zeros(vocabSize, K, 'uint32');
ntheta_ts = zeros(1, K, 'uint32');
ntheta_d = zeros(vocabSize, numDocs, 'uint32');
ntheta_ds = zeros(1, numDocs, 'uint32');
ntheta_n = zeros(vocabSize, 1, 'uint32');
ntheta_ns = zeros(1, 1, 'uint32');

ntau = zeros(K, 1, 'uint32');
ntaus = zeros(1, 1, 'uint32');
nsigma = zeros(3, numDocs, 'uint32');
nsigmas = zeros(1, numDocs, 'uint32');
nvs = zeros(K - 1, 1, 'uint32');


%% ---------------------------------------------------------
% Initialize estimates
% Sentence and word topic assignments
samples(tail).t = uint32(zeros(numDocs, maxNumSents));
samples(tail).s = struct('inner', cell(3, numDocs));


% Initialize v depending on model
samples(tail).v = zeros(K - 1, numDocs, 'uint32');
if model == 2
    % Initialize by Mallows, draw from its prior
    for i = 1:K-1
        % Compute probabilities
        prob = exp(-rhoprior .* [0:1:(K - i)]);
        prob = prob ./ sum(prob);
        
        % Draw samples according to those probabilities
        for j = 1:numDocs
            samples(tail).v(i, j) = uint32(find(mnrnd(1, prob)) - 1);
        end
    end
elseif model == 1
    % Initialize uniformly
    for i = 1:K-1
        samples(tail).v(i, :) = uint32(floor(unifrnd(0, K - i + 1, 1, numDocs)));
    end
end


nvs = uint32(sum(samples(tail).v, 2));
samples(tail).rho = rhoprior * ones(K - 1, 1);
for i = 1:numDocs
    % Sample topic counts
    topicCounts = uint32(sort(ceil(unifrnd(0, K, 1, numSents(i))), 2));
    perm = vToPi(samples(tail).v(:, i));
        
    % Sample word sources
    samples(tail).s(1, i).inner = sparse([], [], [], vocabSize, numSents(i), nzmax(documents(i).count));
    samples(tail).s(2, i).inner = sparse([], [], [], vocabSize, numSents(i), nzmax(documents(i).count));
    samples(tail).s(3, i).inner = sparse([], [], [], vocabSize, numSents(i), nzmax(documents(i).count));
    for j = 1:numSents(i)
        samples(tail).t(i, j) = uint32(perm(topicCounts(j)));
        
        % Update ntau cache
        ntau(samples(tail).t(i, j), 1) = ntau(samples(tail).t(i, j), 1) + 1;
        ntaus(1, 1) = ntaus(1, 1) + 1;
        
        % Valid words for this sentence
        [wordIdx, junk, wordCount] = find(documents(i).count(:, j));
        rands = ceil(unifrnd(0, 3, sum(wordCount), 1));
        idx = 1;
        for k = 1:size(wordIdx)
            for l = 1:wordCount(k)
                source = 3;
 
                samples(tail).s(source, i).inner(wordIdx(k), j) = samples(tail).s(source, i).inner(wordIdx(k), j) + 1;
                if source == 1
                    ntheta_n(wordIdx(k), 1) = ntheta_n(wordIdx(k), 1) + 1;
                    ntheta_ns(1, 1) = ntheta_ns(1, 1) + 1;
                elseif source == 2
                    ntheta_d(wordIdx(k), i) = ntheta_d(wordIdx(k), i) + 1;
                    ntheta_ds(1, i) = ntheta_ds(1, i) + 1;
                else
                    ntheta_t(wordIdx(k), samples(tail).t(i, j)) = ntheta_t(wordIdx(k), samples(tail).t(i, j)) + 1;
                    ntheta_ts(1, samples(tail).t(i, j)) = ntheta_ts(1, samples(tail).t(i, j)) + 1;
                end
                nsigma(source, i) = nsigma(source, i) + 1;
                nsigmas(1, i) = nsigmas(1, i) + 1;
                idx = idx + 1;
            end
        end
    end
end



if savetime == 0
   savetime = iterations;
end
int32(iterations/savetime)
for iter = 1:int32(iterations/savetime)
    prev = tail;
    tail = tail + 1;
    if tail > keep
        tail = 1;
    end
        
    [samples(tail).t ...
     samples(tail).s ...
     samples(tail).v ...
     samples(tail).rho ...
     ntheta_n ...
     ntheta_ns ...
     ntheta_d ...
     ntheta_ds ...
     ntheta_t ...
     ntheta_ts ...
     ntau ...
     ntaus ...
     nsigma ...
     nsigmas ...
     nvs] = ...
            sample_model (documents, ...
                           uint32(numSents), ...
                           samples(prev).t' - 1, ...
                           samples(prev).s, ...
                           samples(prev).v, ...
                           samples(prev).rho, ...
                           theta_n_0, ...
                           theta_d_0, ...
                           theta_t_0, ...
                           tau_0, ...
                           sigma_0, ...
                           v_0, ...
                           nu_0, ...
                           ntheta_n, ...
                           ntheta_ns, ...
                           ntheta_d, ...
                           ntheta_ds, ...
                           ntheta_t, ...
                           ntheta_ts, ...
                           ntau, ...
                           ntaus, ...
                           nsigma, ...
                           nsigmas, ...
                           nvs, ...
                           savetime, ...
                           10, ...
                           model);
                    
    samples(tail).t = samples(tail).t' + 1;

    fprintf('Iteration %f\n', iter * savetime);
    fprintf('Saving intermediate file');
    save(strcat(filename, '.', int2str(iter * savetime), '.mat'), 'samples', 'tail', 'model', 'K', 'lmprior', 'tauprior', 'sigmaprior', 'rhoprior', 'nuprior', 'keep', 'iterations', 'savetime');
end

save(strcat(filename, '.mat'), 'samples', 'tail', 'model', 'K', 'lmprior', 'tauprior', 'sigmaprior', 'rhoprior', 'nuprior', 'keep', 'iterations', 'savetime');
csvwrite(strcat(filename, '.csv'), samples(tail).t);
