#include "mex.h"
#include "matrix.h"
#include "defines.h"
#include <float.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

/* Globals for documents and sentence structures */
mxArray* g_documents;
unsigned int g_totalWords;
unsigned int* g_numSents;
mwSize g_numDocs;
unsigned int g_maxNumSents;
unsigned int g_totalNumSents;
size_t g_vocabSize;
size_t g_K;
int g_model;

/* Globals for samples */
unsigned int* g_t;
mxArray* g_s;
unsigned int* g_v;
double* g_rho;

/* Globals for priors */
double g_theta_n_0;
double g_theta_d_0;
double g_theta_t_0;
double g_tau_0;
double* g_sigma_0;
double g_sigma_0s;
double* g_v_0;
double g_nu_0;

/* Caches */
unsigned int* g_ntheta_n;
unsigned int* g_ntheta_ns;
unsigned int* g_ntheta_d;
unsigned int* g_ntheta_ds;
unsigned int* g_ntheta_t;
unsigned int* g_ntheta_ts;
unsigned int* g_ntau;
unsigned int* g_ntaus;
unsigned int* g_nsigma;
unsigned int* g_nsigmas;
unsigned int* g_nvs;

/* Random values from MATLAB */
double* g_unifrand;
unsigned int g_randIdx;

/* For t and v resamplers */
/* Cache for topic probabilities, indexed by span and topic; first index is topic,
   second is start of topic run, third index is length (span) of run */
double* g_sentProbCache;
/* Inversion switch probabilities */
double* g_vProb;
/* Topic switch probabilities */
double* g_topicProb;
/* Proposed new v's, t's, and pi's */
unsigned int* g_proposedV;
unsigned int* g_proposedOrdering;
unsigned int* g_proposedTopics;
/* Counts within a document */
unsigned int* g_documentWords;

/* Helper files */
#include "utility.c"
#include "permutation.c"
#include "sentenceProb.c"
#include "resample_t.c"
#include "resample_v.c"
#include "resample_rho.c"

/* Adds/removes a document from topic word caches */
void modifyNTheta_T(unsigned int document, bool add)
{
	mxArray* words = mxGetFieldByNumber(g_s, index2d(2, 3, document), 0);
	mwIndex* ir = mxGetIr(words);
	mwIndex* jc = mxGetJc(words);
	double* pr = mxGetPr(words);
	unsigned int* topics = g_t + index2d(0, g_maxNumSents, document);

	unsigned int i, j;

	for (i = 0; i < g_numSents[document]; ++i)
	{
		for (j = (unsigned int)jc[i]; j < jc[i + 1]; ++j)
		{
			if (add == true)
			{
				g_ntheta_t[index2d(ir[j], g_vocabSize, topics[i])] += round(pr[j]);
				g_ntheta_ts[index2d(0, 1, topics[i])] += round(pr[j]);
			}
			else
			{
				g_ntheta_t[index2d(ir[j], g_vocabSize, topics[i])] -= round(pr[j]);
				g_ntheta_ts[index2d(0, 1, topics[i])] -= round(pr[j]);
			}
		}
	}
}

int initialize(int lhsCount, mxArray* lhs[], int rhsCount, const mxArray* rhs[])
{
	/* Documents */
	g_documents = (mxArray*)rhs[R_DOCUMENTS];
	g_numSents = (unsigned int*)mxGetData(rhs[R_NUMSENTS]);
	g_numDocs = mxGetM(rhs[R_NUMSENTS]);
	g_maxNumSents = findmax(g_numSents, g_numDocs);
	g_totalNumSents = sum(g_numSents, g_numDocs);
	g_vocabSize = mxGetM(rhs[R_NTHETA_N]);
	g_K = mxGetN(rhs[R_NTHETA_T]);
	g_model = (int)(mxGetScalar(rhs[R_MODEL]) + 0.5);

	/* Samples */
	g_t = (unsigned int*)mxGetData(lhs[L_T] = mxDuplicateArray(rhs[R_T]));
	g_s = lhs[L_S] = spaceOut(g_numDocs, g_numSents, rhs[R_S], g_documents, &g_totalWords);
	g_v = (unsigned int*)mxGetData(lhs[L_V] = mxDuplicateArray(rhs[R_V]));
	g_rho = mxGetPr(lhs[L_RHO] = mxDuplicateArray(rhs[R_RHO]));

	/* Priors */
	g_theta_n_0 = mxGetScalar(rhs[R_THETA_N_0]);
	g_theta_d_0 = mxGetScalar(rhs[R_THETA_D_0]);
	g_theta_t_0 = mxGetScalar(rhs[R_THETA_T_0]);
	g_tau_0 = mxGetScalar(rhs[R_TAU_0]);
	g_sigma_0 = mxGetPr(rhs[R_SIGMA_0]);
	g_sigma_0s = g_sigma_0[0] + g_sigma_0[1] + g_sigma_0[2];
	g_v_0 = mxGetPr(rhs[R_V_0]);
	g_nu_0 = mxGetScalar(rhs[R_NU_0]);

	/* Caches */
	g_ntheta_n = (unsigned int*)mxGetData(lhs[L_NTHETA_N] = mxDuplicateArray(rhs[R_NTHETA_N]));
	g_ntheta_ns = (unsigned int*)mxGetData(lhs[L_NTHETA_NS] = mxDuplicateArray(rhs[R_NTHETA_NS]));

	g_ntheta_d = (unsigned int*)mxGetData(lhs[L_NTHETA_D] = mxDuplicateArray(rhs[R_NTHETA_D]));
	g_ntheta_ds = (unsigned int*)mxGetData(lhs[L_NTHETA_DS] = mxDuplicateArray(rhs[R_NTHETA_DS]));

	g_ntheta_t = (unsigned int*)mxGetData(lhs[L_NTHETA_T] = mxDuplicateArray(rhs[R_NTHETA_T]));
	g_ntheta_ts = (unsigned int*)mxGetData(lhs[L_NTHETA_TS] = mxDuplicateArray(rhs[R_NTHETA_TS]));

	g_ntau = (unsigned int*)mxGetData(lhs[L_NTAU] = mxDuplicateArray(rhs[R_NTAU]));
	g_ntaus = (unsigned int*)mxGetData(lhs[L_NTAUS] = mxDuplicateArray(rhs[R_NTAUS]));

	g_nsigma = (unsigned int*)mxGetData(lhs[L_NSIGMA] = mxDuplicateArray(rhs[R_NSIGMA]));
	g_nsigmas = (unsigned int*)mxGetData(lhs[L_NSIGMAS] = mxDuplicateArray(rhs[R_NSIGMAS]));

	g_nvs = (unsigned int*)mxGetData(lhs[L_NVS] = mxDuplicateArray(rhs[R_NVS]));

	/* Other variables for resamplers */
	g_sentProbCache = mxMalloc(g_maxNumSents * g_maxNumSents * g_K * sizeof(double));
	g_vProb = mxMalloc(g_K * sizeof(double));
	g_topicProb = mxMalloc(g_K * sizeof(double));
	g_proposedV = mxMalloc((g_K - 1) * sizeof(unsigned int));
	g_proposedOrdering = mxMalloc(g_K * sizeof(unsigned int));
	g_proposedTopics = mxMalloc(g_maxNumSents * sizeof(unsigned int));
	g_documentWords = mxMalloc(g_vocabSize * sizeof(unsigned int));

	return 0;
}

void mexFunction(int lhsCount, mxArray* lhs[], int rhsCount, const mxArray* rhs[])
{
	/* Extremely stupid useless variable declaration - used to allow initialize to be called up front under C89 convention */
	int init = initialize(lhsCount, lhs, rhsCount, rhs);

	/* Stuff specific to this run of the sampler */
	unsigned int iterations = round(mxGetScalar(rhs[R_ITERATIONS]));
	unsigned int rhoBurnin = round(mxGetScalar(rhs[R_RHO_BURNIN]));

	unsigned int iters;
	unsigned int i, j, k, l, m, n;

	/* Random numbers */
	/* Stuff for calling unifrnd */
	mxArray* unifrandArray;
	mxArray* zero = mxCreateDoubleScalar(0.0);
	mxArray* one = mxCreateDoubleScalar(1.0);
	mxArray* randLength = mxCreateDoubleScalar(0.0);
	double* randLengthValue = mxGetPr(randLength);
	mxArray* unifrndParams[4] = { zero, one, randLength, one };

	/* Random permutation */
	mxArray* orderArray;
	double* order;
	mxArray* permLength = mxCreateDoubleScalar(0.0);
	double* permLengthData = mxGetPr(permLength);

	/* RHO SAMPLING */

	/* Stuff for calling our outer custom rho sampler */
	mxArray* totalVArray = mxCreateDoubleScalar(0.0);
	double* totalVParam = mxGetPr(totalVArray);
	mxArray* numDocsArray = mxCreateDoubleScalar((double)g_numDocs);
	mxArray* nminusjArray = mxCreateDoubleScalar(0.0);
	double* nminusjParam = mxGetPr(nminusjArray);
	mxArray* v_0Array = mxCreateDoubleScalar(0.0);
	double* v_0Param = mxGetPr(v_0Array);
	mxArray* nu_0Array = mxCreateDoubleScalar(g_nu_0);
	mxArray* lastRhoArray = mxCreateDoubleScalar(0.0);
	double* lastRhoParam = mxGetPr(lastRhoArray);
	mxArray* burninArray = mxCreateDoubleScalar((double)rhoBurnin);
	mxArray* sample_rhoParams[7] = { totalVArray, numDocsArray, nminusjArray, v_0Array, nu_0Array, lastRhoArray, burninArray };

	int status;
	
	/* END RHO SAMPLING */

	g_inversePermutation = mxMalloc(g_K * sizeof(unsigned int));

	mexSetTrapFlag(1);

	/* Reorder topics to correct order a priori */
	for (i = 0; i < g_numDocs; ++i)
	{
		setGlobalPermutation(g_v + index2d(0, g_K - 1, i));
		qsort(g_t + index2d(0, g_maxNumSents, i), g_numSents[i], sizeof(unsigned int), permutationCompare);
	}

	for (iters = 0; iters < iterations; ++iters)
	{
		mexCallMATLAB(0, NULL, 0, NULL, "tic");

		/* Random numbers for this iteration */
		*randLengthValue = (double)g_totalNumSents + (double)g_numDocs * (double)(g_K - 1);
		mexCallMATLAB(1, &unifrandArray, 4, unifrndParams, "unifrnd"); 
		g_unifrand = mxGetPr(unifrandArray);
		g_randIdx = 0;

		for (i = 0; i < g_numDocs; ++i)
		{
			/* Remove all values of this document from LM cache ntheta_t */
			modifyNTheta_T(i, false);

			/* Clear out appropriate amount of space in cache for this document */
			memset(g_sentProbCache, 0, g_numSents[i] * g_numSents[i] * g_K * sizeof(double));

			/* RESAMPLE t */

			/* Generate random order of topics to resample */
			*permLengthData = (double)g_numSents[i];
			mexCallMATLAB(1, &orderArray, 1, &permLength, "randperm");
			order = mxGetPr(orderArray);

			resample_t(i, order);

			/* Clean up */
			mxDestroyArray(orderArray);

			/* RESAMPLE v */

			if (g_model == MALLOWS_LDA || g_model == UNIFORM_LDA)
			{
				/* Resample each position in V vector */
				*permLengthData = (double)(g_K - 1);
				mexCallMATLAB(1, &orderArray, 1, &permLength, "randperm");
				order = mxGetPr(orderArray);

				resample_v(i, order);

				mxDestroyArray(orderArray);
			}

			/* Resort actual topics */
			setGlobalPermutation(g_v + index2d(0, g_K - 1, i));
			qsort(g_t + index2d(0, g_maxNumSents, i), g_numSents[i], sizeof(unsigned int), permutationCompare);

			/* Add back new topics to LM cache ntheta_t */
			modifyNTheta_T(i, true);

			/* RESAMPLE c */

		}

		/* RESAMPLE theta */

		/* RESAMPLE Phi */

		/* RESAMPLE Sigma */

		/* RESAMPLE rho */
		if (g_model == MALLOWS_LDA)
		{
			resample_rho(totalVParam, nminusjParam, v_0Param, lastRhoParam, sample_rhoParams);
		}

		/* A little cleanup */
		mxDestroyArray(unifrandArray);

		mexCallMATLAB(0, NULL, 0, NULL, "toc");
		mexPrintf("Iteration: %u of current call\n", iters);
	}

	/* Destroy MATLAB callback arrays */
	mxDestroyArray(permLength);
	mxDestroyArray(zero);
	mxDestroyArray(one);
	mxDestroyArray(randLength);
	mxDestroyArray(totalVArray);
	mxDestroyArray(numDocsArray);
	mxDestroyArray(nminusjArray);
	mxDestroyArray(v_0Array);
	mxDestroyArray(nu_0Array);
	mxDestroyArray(lastRhoArray);
	mxDestroyArray(burninArray);

	/* Destroy locally malloced arrays */
	mxFree(g_proposedV);
	mxFree(g_proposedOrdering);
	mxFree(g_vProb);
	mxFree(g_inversePermutation);
	mxFree(g_topicProb);
	mxFree(g_sentProbCache);
	mxFree(g_proposedTopics);
	mxFree(g_documentWords);
}
