/* Utility functions for modeling */

#include "mex.h"
#include "matrix.h"
#include <math.h>

/* Returns index into sentProbCache specifically (which is a 3D array with first two sizes the same) */
inline size_t cacheIndex(size_t start, size_t span, size_t numSents, size_t topic)
{
	return topic * numSents * numSents + start * numSents + span;
}

/* Returns index into 2D array; matches MATLAB convention */
inline size_t index2d(size_t row, size_t numRows, size_t col)
{
	return col * numRows + row;
}

unsigned int findmax(unsigned int* data, size_t size)
{
	unsigned int i, value = 0;
	for (i = 0; i < size; ++i)
	{
		/* Find maxNumSents */
		if (data[i] > value)
		{
			value = data[i];
		}
	}
	return value;
}

unsigned int sum(unsigned int* data, size_t size)
{
	unsigned int i, value = 0;
	for (i = 0; i < size; ++i)
	{
		value += data[i];
	}
	return value;
}

/* Samples from normalized multinomial using vector of log probs */
unsigned int sampleNormalized(double* logProb, size_t count, double rand)
{
	double sum = 0.0;
	unsigned int i;
	double scaledRandom;
	double maxLogProb = -DBL_MAX;
	for (i = 0; i < count; ++i)
	{
		if (logProb[i] > maxLogProb)
		{
			maxLogProb = logProb[i];
		}
	}
	for (i = 0; i < count; ++i)
	{
		logProb[i] = exp(logProb[i] - maxLogProb);
		sum += logProb[i];
	}
	scaledRandom = rand * sum;
	sum = 0.0;
	for (i = 0; i < count; ++i)
	{
		sum += logProb[i];
		if (scaledRandom < sum)
		{
			break;
		}
	}
	/* Special case */
	if (i == count)
	{
		i = round(floor(rand * count));
	}
	return i;
}

/* Readjusts sparse matrix positioning in s matrix to correspond to document sparse matrix structure */
mxArray* spaceOut(size_t numDocs, unsigned int* numSents, const mxArray* old_s, const mxArray* documents, unsigned int* totalWords)
{
	/* Declare space for new s */
	mxArray* s = mxDuplicateArray(old_s);

	mxArray* doc;
	mwIndex* docIr;
	mwIndex* docJc;

	mxArray* old_s_doc;
	mwIndex* old_sIr;
	mwIndex* old_sJc;
	double* old_sPr;
	mwIndex old_sIrIdx;

	mxArray* s_doc;
	mwIndex* sIr;
	mwIndex* sJc;
	double* sPr;

	size_t nzmax;

	unsigned int i, j, k;
	size_t l;

	*totalWords = 0;

	/* Iterate through documents and assign to new s according to values in old s */
	for (i = 0; i < numDocs; ++i)
	{
		doc = mxGetFieldByNumber(documents, i, 0);
		docIr = mxGetIr(doc);
		docJc = mxGetJc(doc);
		nzmax = mxGetNzmax(doc);
		for (j = 0; j < 3; ++j)
		{
			old_s_doc = mxGetFieldByNumber(old_s, index2d(j, 3, i), 0);
			old_sIr = mxGetIr(old_s_doc);
			old_sJc = mxGetJc(old_s_doc);
			old_sPr = mxGetPr(old_s_doc);
			old_sIrIdx = 0;

			s_doc = mxGetFieldByNumber(s, index2d(j, 3, i), 0);
			sIr = mxGetIr(s_doc);
			sJc = mxGetJc(s_doc);
			sPr = mxGetPr(s_doc);

			/* Copy Ir and Jc into s's */
			memcpy(sIr, docIr, nzmax * sizeof(mwIndex));
			memcpy(sJc, docJc, (numSents[i] + 1) * sizeof(mwIndex));
			memset(sPr, 0, nzmax * sizeof(double));

			/* Copy counts based on previous s to the new s */
			for (k = 0; k < numSents[i]; ++k)
			{
				old_sIrIdx = old_sJc[k];
				/* For each sentence copy Pr items */
				for (l = sJc[k]; l < sJc[k + 1]; ++l)
				{
					if (old_sIrIdx != old_sJc[k + 1] && sIr[l] == old_sIr[old_sIrIdx])
					{
						sPr[l] = old_sPr[old_sIrIdx++];
						*totalWords += round(sPr[l]);
					}
					else
					{
						sPr[l] = 0.0;
					}
				}
			}
		}
	}

	return s;
}

/* Very simple log1p function since Windows doesn't have one */
#ifndef __GNUC__
inline double log1p(double x)
{
	if (x < 1e-6)
	{
		return x;
	}
	else
	{
		return log(x + 1.0);
	}
}
#endif
